/*
 * timeout.c - scheduler timers
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <assert.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "prioq.h"
#include "failures.h"
#include "io.h"
#include "timeout.h"

#define ONE_SECOND		1000
#define USECS_IN_MSEC		1000
#define MSEC_BACKWARD_TOLERANCE	5000

/* Timeouts are global and every active io context runs them */

/* global timeout queue */
static struct prioq *timeout_queue;

/* pointer to timeout who's callback is running now */
static struct timeout *timeout_running = NULL;

/* timestamp of the last timeout_run  */
static struct timeval last_tv;

static int timeout_compare(struct prioq_head *a, struct prioq_head *b)
{
	struct timeout *aptr = (struct timeout *) a;
	struct timeout *bptr = (struct timeout *) b;

	if (aptr->when.tv_sec > bptr->when.tv_sec)
		return -1;
	else if (aptr->when.tv_sec < bptr->when.tv_sec)
		return 1;
	else if (aptr->when.tv_usec > bptr->when.tv_usec)
		return -1;
	else if (aptr->when.tv_usec < bptr->when.tv_usec)
		return 1;
	else if ((unsigned long) aptr > (unsigned long) bptr)
		return -1;
	else if ((unsigned long) aptr < (unsigned long) bptr)
		return 1;
	return 0;
}

static void check_clock_backward(struct timeval *now)
{
	int msec;
	struct timeval res;

	if (last_tv.tv_sec > 0) {
		res.tv_sec = now->tv_sec - last_tv.tv_sec;
		res.tv_usec = now->tv_usec - last_tv.tv_usec;
		if (res.tv_usec < 0) {
			res.tv_sec -= 1;
			res.tv_usec += 1000000;
		}
		msec = res.tv_sec * 1000 + res.tv_usec / 1000;
		if (msec < 0 && msec < -MSEC_BACKWARD_TOLERANCE)
			xerror(1, "timeout: timer clock runs backward too long and i don't know what to and i terminate.");
	}
	last_tv = *now;
}

struct timeout *timeout_new(iocallback_t callback, void *opaque)
{
	struct timeout *tout;

	if (!callback)
		xerror(1, "timeout_new: callback NULL");

	tout = xmalloc0(sizeof(*tout));
	tout->qh.index = -1;
	tout->opaque = opaque;
	tout->when.tv_sec = 0;
	tout->when.tv_usec = 0;
	tout->callback = callback;
	tout->msec = -1;
	return tout;
}

int timeout_active(struct timeout *tout)
{
	if (!tout)
		xerror(1, "timeout_active: tout NULL");

	if (tout->qh.index >= 0)
		return 1;

	return 0;
}

void timeout_free(struct timeout **tout)
{
	if (!tout || *tout == NULL)
		xerror(1, "timeout_free: tout NULL");

	if ((*tout)->qh.index >= 0)
		xerror(1, "timeout_free: attempt to free active timer");

	xfree(*tout);
	*tout = NULL;
}

void timeout_add(struct timeout *tout, int msec)
{
	struct timeval now;

	if (!tout)
		xerror(1, "timeout_add: args NULL");

	if (tout->qh.index >= 0)
		xerror(1, "timeout_add: timer %p already in queue", tout);

	if (gettimeofday(&now, NULL))
		xwarning("timeout_add: gettimeofday()");

	now.tv_sec += msec / 1000;
	now.tv_usec += (msec % 1000) * 1000;
	if (now.tv_usec >= 1000000) {
		now.tv_sec += 1;
		now.tv_usec -= 1000000;
	}
	tout->msec = msec;
	tout->when = now;
	prioq_put(timeout_queue, (struct prioq_head *) tout);
}

void timeout_del(struct timeout *tout)
{
	if (!tout)
		xerror(1, "timeout_del: tout NULL");

	if (tout->qh.index >= 0) {
		if (timeout_running && timeout_running == tout) {
			xwarning("timeout_del: attempt to delete timeout %p inside it's callback", tout);
			return;
		}
		prioq_remove(timeout_queue, (struct prioq_head *)tout);
		return;
	}
	xwarning("timeout_del: attempt to delete inactive timer %p", tout);
}

int timeout_next(void)
{
	int msecs = 0;
	struct timeval now;
	struct timeout *tout;

	if (gettimeofday(&now, NULL))
		xwarning("timeout_next: gettimeofday()");

	if ((tout = (struct timeout *) prioq_peek(timeout_queue)) != NULL) {
		now.tv_sec = tout->when.tv_sec - now.tv_sec;
		now.tv_usec = tout->when.tv_usec - now.tv_usec;
		if (now.tv_usec < 0) {
			now.tv_sec -= 1;
			now.tv_usec += 1000000;
		}
		if (now.tv_sec < 0) {
			now.tv_sec = 0;
			now.tv_usec = 0;
		}
		msecs = now.tv_sec * 1000;
		msecs += (now.tv_usec + USECS_IN_MSEC) / 1000;
		return msecs > ONE_SECOND ? ONE_SECOND : msecs;
	}

	return ONE_SECOND;
}

int timeout_run(void)
{
	int count = 0;
	struct timeout *tout;
	struct timeval now;

	if (gettimeofday(&now, NULL))
		xwarning("timeout_run: gettimeofday()");

	check_clock_backward(&now);

	while ((tout = (struct timeout *) prioq_peek(timeout_queue)) != NULL) {
		if (tout->when.tv_sec < now.tv_sec ||
		    (tout->when.tv_sec == now.tv_sec &&
		     (tout->when.tv_usec / USECS_IN_MSEC) <
		      (now.tv_usec / USECS_IN_MSEC))) {
			prioq_remove(timeout_queue, (struct prioq_head *) tout);
			assert(tout != NULL);
			assert(tout->callback != NULL);
			timeout_running = tout;
			(*tout->callback)(-1, EVENT_TIMEOUT, tout->opaque);
			timeout_running = NULL;
			count++;
		} else
			break;
	}
	return count;
}

void timeout_init(void)
{
	if (timeout_queue) {
		xwarning("timeout_init: already initialized");
		return;
	}

	timeout_queue = prioq_new(timeout_compare);
	timeout_running = NULL;
	last_tv.tv_sec = 0;
	last_tv.tv_usec = 0;
}

void timeout_deinit(void)
{
	if (!timeout_queue) {
		xwarning("timeout_deinit: not initialized");
		return;
	}

	timeout_running = NULL;
	last_tv.tv_sec = 0;
	last_tv.tv_usec = 0;
	prioq_free(&timeout_queue);
}

